{% from "macros.html" import message_expansion %}

<div class="message-log" id="message-log-{{ log.id }}">
    {% for message in log.message_log %}
        {{ message_expansion(message, loop.index, expanded) }}
    {% endfor %}
</div>
